// (c) 2010 OpenSeadragon, (c) 2010 CodePlex Foundation
/// <reference path="embed/OpenSeadragon.pre.js" />

Array.prototype.add = function(array, item) {
    array[array.length] = item;
};

Array.prototype.clear = function(array) {
    array.length = 0;
};

Array.prototype.clone = function(array) {
    return array.length === 1 ? [array[0]] : Array.apply(null, array);
};
SArray = Array();

if (!window.Function) {
    window.Function = {};
}
Function.prototype.createDelegate = function (object, method) {
    return function () {
        if (arguments === undefined)
            arguments = [];
        return method.apply(object, arguments);
    };
};

function String() {

}
String.format = function (format, args) {
    return String._toFormattedString(false, arguments);
};
String._toFormattedString = function (useLocale, args) {
    var result = '';
    var format = args[0];

    for (var i = 0; ; ) {
        var open = format.indexOf('{', i);
        var close = format.indexOf('}', i);
        if ((open < 0) && (close < 0)) {
            result += format.slice(i);
            break;
        }
        if ((close > 0) && ((close < open) || (open < 0))) {
            if (format.charAt(close + 1) !== '}') {
                throw Error.argument('format', Sys.Res.stringFormatBraceMismatch);
            }
            result += format.slice(i, close + 1);
            i = close + 2;
            continue;
        }

        result += format.slice(i, open);
        i = open + 1;

        if (format.charAt(i) === '{') {
            result += '{';
            i++;
            continue;
        }

        if (close < 0) throw Error.argument('format', Sys.Res.stringFormatBraceMismatch);


        var brace = format.substring(i, close);
        var colonIndex = brace.indexOf(':');
        var argNumber = parseInt((colonIndex < 0) ? brace : brace.substring(0, colonIndex), 10) + 1;
        if (isNaN(argNumber)) throw Error.argument('format', Sys.Res.stringFormatInvalid);
        var argFormat = (colonIndex < 0) ? '' : brace.substring(colonIndex + 1);

        var arg = args[argNumber];
        if (typeof (arg) === "undefined" || arg === null) {
            arg = '';
        }

        if (arg.toFormattedString) {
            result += arg.toFormattedString(argFormat);
        }
        else if (useLocale && arg.localeFormat) {
            result += arg.localeFormat(argFormat);
        }
        else if (arg.format) {
            result += arg.format(argFormat);
        }
        else
            result += arg.toString();

        i = close + 1;
    }

    return result;
};

function Observer() { }
Observer.prototype = {
    _getContext: function (obj, create) {
        var ctx = obj._observerContext;
        if (ctx) return ctx();
        if (create) {
            return (obj._observerContext = this._createContext())();
        }
        return null;
    },
    _createContext: function () {
        var ctx = {
            events: new EventHandlerList()
        };
        return function () {
            return ctx;
        };
    }
};

var EventHandlerList = function EventHandlerList() {
    this._list = {};
};
EventHandlerList.prototype = {
    _addHandler: function (id, handler) {
        SArray.add(this._getEvent(id, true), handler);
    },
    addHandler: function (id, handler) {
        this._addHandler(id, handler);
    },
    _removeHandler: function (id, handler) {
        var evt = this._getEvent(id);
        if (!evt) return;
        SArray.remove(evt, handler);
    },
    _removeHandlers: function (id) {
        if (!id) {
            this._list = {};
        }
        else {
            var evt = this._getEvent(id);
            if (!evt) return;
            evt.length = 0;
        }
    },
    removeHandler: function (id, handler) {
        this._removeHandler(id, handler);
    },
    getHandler: function (id) {
        var evt = this._getEvent(id);
        if (!evt || !evt.length) return null;
        evt = SArray.clone(evt);
        return function (source, args) {
            for (var i = 0, l = evt.length; i < l; i++) {
                evt[i](source, args);
            }
        };
    },
    _getEvent: function (id, create) {
        var e = this._list[id];
        if (!e) {
            if (!create) return null;
            this._list[id] = e = [];
        }
        return e;
    }
};
